"""
Functions ported from the R package sm.

Implements different bandwidth selection methods, including:
- Scott's rule of thumb
- Silverman's rule of thumb
- Sheather-Jones estimator
- GridSearchCV best estimator
"""
import numpy as np
from scipy.stats import norm
import scipy.interpolate as interpolate
from pandas import Series
from sklearn.model_selection import GridSearchCV
from sklearn.neighbors import KernelDensity
from typing import Callable, Union
import warnings

from xnb.enums import BWFunctionName as BFN

__all__ = [
    'get_bandwidth_function',
    'hsilverman',
    'hscott',
    'hsj',
    'best_estimator',
]


def get_bandwidth_function(bf: str) -> Callable:
  """Return a bandwidth function based on the given BWFunctionName.

  Example of use:
  >>> from xnb.enum import BWFunctionName
  >>> from xnb._bandwidth_functions import get_bandwidth_function
  >>> bw_function = get_bandwidth_function(BWFunctionName.HSILVERMAN)
  >>> bw_function(data)
  """
  bw_functions_dict = {
      BFN.HSILVERMAN: _bandwidth_wrapper(hsilverman),
      BFN.HSCOTT: _bandwidth_wrapper(hscott),
      BFN.HSJ: _bandwidth_wrapper(hsj),
      BFN.BEST_ESTIMATOR: _bandwidth_wrapper(best_estimator)
  }
  return bw_functions_dict[bf]


def _bandwidth_wrapper(bw_function):
  def _bw_wrapper(data, n_sample):
    if bw_function.__name__ == BFN.BEST_ESTIMATOR.value:
      return bw_function(data, n_sample)
    return bw_function(data)
  return _bw_wrapper


def hsilverman(data: Series) -> float:
  """Silverman's rule of thumb.

  This approximation is termed the normal distribution approximation,
  Gaussian approximation, or Silverman's rule of thumb.
  While this rule of thumb is easy to compute, it should be used with caution
  as it can yield widely inaccurate estimates when the density is not close
  to being normal.
  """
  A = _select_sigma(data)
  n = len(data)
  return .9 * A * n ** (-0.2)


def hscott(data: Series) -> float:
  """Scott's rule of thumb.

  Use Scott's rule of thumb to determine the smoothing bandwidth for the
  kernel estimation of point process intensity.
  """
  A = _select_sigma(data)
  n = len(data)
  return 1.059 * A * n ** (-0.2)


def hsj(data: Series) -> float:
  """Sheather-Jones bandwidth estimator.

    References
  ----------
    A reliable data-based bandwidth selection method for kernel
    density estimation. Simon J. Sheather and Michael C. Jones.
    Journal of the Royal Statistical Society, Series B. 1991
  """
  h0 = _hnorm(data)
  v0 = _sj(data, h0)

  if v0 > 0:
    hstep = 1.1
  else:
    hstep = 0.9

  h1 = h0 * hstep
  v1 = _sj(data, h1)

  while v1 * v0 > 0:
    h0 = h1
    v0 = v1
    h1 = h0 * hstep
    v1 = _sj(data, h1)

  result = h0 + (h1 - h0) * abs(v0) / (abs(v0) + abs(v1))
  if np.isnan(result):
    warnings.warn(
        'Bandwidth - HSJ - The HSJ function was going to return '
        'NaN due to the nature of the data. To prevent errors, '
        '0.0000001 for this feature and class will be returned instead. '
        f'\nData:\n {data}', UserWarning, stacklevel=2
    )
    return 0.0000001
  else:
    return result


def best_estimator(data: Series, n_sample: int) -> float:
  """Return the best estimator of the bandwidth using GridSearchCV."""
  rang = abs(max(data) - min(data))
  if rang == 0:
    warnings.warn(
        'Bandwidth - Best Estimator - There is no deviation '
        'in one of the features provided for a specific class.'
        ' The bandwidth result will be 0.0000001 for this feature and class.'
        f'\nData:\n {data}', UserWarning, stacklevel=2
    )
    return 0.0000001
  len_unique = len(data.unique())
  params = {
      'bandwidth': np.linspace(
          start=rang / len_unique,
          stop=rang,
          num=n_sample
      )
  }
  data_ = data.values[:, np.newaxis]
  grid = GridSearchCV(KernelDensity(), params, cv=3)
  grid.fit(list(data_))
  return float(grid.best_estimator_.bandwidth_)


def _hnorm(data: Series, weights: Union[np.ndarray, None] = None) -> float:

  x = np.asarray(data)

  if weights is None:
    weights = np.ones(len(x))

  n = float(sum(weights))

  if len(x.shape) == 1:
    sd = np.sqrt(_wvar(x, weights))
    return sd * (4 / (3 * n)) ** (1 / 5.0)

  # TODO: make this work for more dimensions
  # ((4 / (p + 2) * n)^(1 / (p+4)) * sigma_i
  if len(x.shape) == 2:
    ndim = x.shape[1]
    sd = np.sqrt(np.apply_along_axis(_wvar, 1, x, weights))
    return (4.0 / ((ndim + 2.0) * n) ** (1.0 / (ndim + 4.0))) * sd

  return 0.0


def _sj(x, h):
  def phi6(x):
    return (x ** 6 - 15 * x ** 4 + 45 * x ** 2 - 15) * _dnorm(x)

  def phi4(x):
    return (x ** 4 - 6 * x ** 2 + 3) * _dnorm(x)

  n = len(x)
  one = np.ones((1, n))

  lam = np.percentile(x, 75) - np.percentile(x, 25)
  a = 0.92 * lam * n ** (-1 / 7.0)
  b = 0.912 * lam * n ** (-1 / 9.0)

  W = np.tile(x, (n, 1))
  W = W - W.T

  W1 = phi6(W / b)
  tdb = np.dot(np.dot(one, W1), one.T)
  tdb = -tdb / (n * (n - 1) * b ** 7)

  W1 = phi4(W / a)
  sda = np.dot(np.dot(one, W1), one.T)
  sda = sda / (n * (n - 1) * a ** 5)

  alpha2 = 1.357 * (abs(sda / tdb)) ** (1 / 7.0) * h ** (5 / 7.0)

  W1 = phi4(W / alpha2)
  sdalpha2 = np.dot(np.dot(one, W1), one.T)
  sdalpha2 = sdalpha2 / (n * (n - 1) * alpha2 ** 5)

  return (
      norm.pdf(0, 0, np.sqrt(2)) / (n * abs(sdalpha2[0, 0]))
  ) ** 0.2 - h


def _select_sigma(x):
  """Return the smaller of std(X, ddof=1) or normalized IQR(X) over axis 0."""
  normalize = 1.349
  IQR = (_scoreatpercentile(x, 75) - _scoreatpercentile(x, 25)) / normalize
  std_dev = np.std(x, axis=0, ddof=1)
  if IQR > 0:
    return np.minimum(std_dev, IQR)
  elif std_dev > 0:
    return std_dev
  else:
    return 0.0000001


def _scoreatpercentile(data, percentile):
  """Return the score at the given percentile of the data."""
  per = np.array(percentile)
  cdf = _empiricalcdf(data)
  interpolator = interpolate.interp1d(np.sort(cdf), np.sort(data))
  return interpolator(per / 100.)


def _empiricalcdf(data: np.ndarray, method: str = 'Hazen') -> np.ndarray:
  """Return the empirical cdf.

  Methods available:
      Hazen:       (i-0.5)/N
      Weibull:     i/(N+1)
      Chegodayev:  (i-.3)/(N+.4)
      Cunnane:     (i-.4)/(N+.2)
      Gringorten:  (i-.44)/(N+.12)
      California:  (i-1)/N
  Where i goes from 1 to N.

  # Args:
      data: 1-D array
      method: string, default 'Hazen'
  # Returns:
      cdf: 1-D array
  """
  i = np.argsort(np.argsort(data)) + 1.
  N = len(data)
  method = method.lower()
  # TODO: Is it worth having all these options?
  if method == 'hazen':
    cdf = (i - 0.5) / N
  elif method == 'weibull':
    cdf = i / (N + 1.)
  elif method == 'california':
    cdf = (i - 1.) / N
  elif method == 'chegodayev':
    cdf = (i - .3) / (N + .4)
  elif method == 'cunnane':
    cdf = (i - .4) / (N + .2)
  elif method == 'gringorten':
    cdf = (i - .44) / (N + .12)
  else:
    raise ValueError('Unknown method. Choose among Weibull, Hazen,'
                     'Chegodayev, Cunnane, Gringorten and California.')

  return cdf


def _wmean(x, w):
  """Weighted mean."""
  return sum(x * w) / float(sum(w))


def _wvar(x, w):
  """Weighted variance."""
  return sum(w * (x - _wmean(x, w)) ** 2) / float(sum(w) - 1)


def _dnorm(x):
  return norm.pdf(x, 0.0, 1.0)



"""Explicable Naive Bayes."""
from sklearn.neighbors import KernelDensity
from typing import Tuple, List, Dict, Set, Union
from pandas import DataFrame, Series
from collections import defaultdict
from itertools import product
from math import sqrt
import numpy as np
from sklearn.base import BaseEstimator

from xnb._kde_object import KDE
from xnb.enums import Algorithm, BWFunctionName, Kernel
from xnb._bandwidth_functions import get_bandwidth_function
from xnb._progress_bar import progress_bar


__all__ = [
    'XNB',
    'NotFittedError'
]


class _ClassFeatureDistance:
  def __init__(self, class_value: str, feature: str, distance: float = 0.0):
    self.class_value = class_value
    self.feature = feature
    self.distance = distance

  def __hash__(self):
    return hash(str(self.class_value) + str(self.feature))


class XNB(BaseEstimator):
  """Class to perform Explainable Naive Bayes."""

  def __init__(
      self,
      bw_function: Union[BWFunctionName, str] = BWFunctionName.HSILVERMAN,
      kernel: Union[Kernel, str] = Kernel.GAUSSIAN,
      algorithm: Union[Algorithm, str] = Algorithm.AUTO,
      n_sample: int = 50,
      show_progress_bar: bool = False
  ) -> None:
    """Initialize the Explainable Naive Bayes model.

    Args:
        bw_function (Union[BWFunctionName, str], optional): Bandwidth selection
          function. Defaults to BWFunctionName.HSILVERMAN.
        kernel (Union[Kernel, str], optional): Kernel type for density
          estimation. Defaults to Kernel.GAUSSIAN.
        algorithm (Union[Algorithm, str], optional): Algorithm for kernel
          density estimation. Defaults to Algorithm.AUTO.
        n_sample (int, optional): Number of samples for density estimation.
          Defaults to 50.
        show_progress_bar (bool, optional): Whether to display a progress bar
          during fitting. Defaults to False.
    """
    self.bw_function = str(bw_function)
    self.kernel = str(kernel)
    self.algorithm = str(algorithm)
    self.n_sample = n_sample
    self.show_progress_bar = show_progress_bar

  @property
  def feature_selection_dict(
      self
  ) -> Dict[Union[str, float], Set[Union[str, float]]]:
    """Obtain the feature selection dictionary.

    Each target class (key) is associated with a set of features (values).
    """
    if hasattr(self, '_feature_selection_dict'):
      return self._feature_selection_dict
    else:
      raise NotFittedError()

  def _repr_html_(self):
    """Return html representation of the model."""
    return ('''
      <style>
          .sklearn-mime-model {{
              display: inline-block;
              padding: 8px 12px;
              background: #f8f9fa;
              border-left: 5px solid #007bff;
              border-radius: 4px;
              font-family: monospace;
              font-size: 14px;
          }}  # noqa
      </style>
    '''
    )  # noqa

  def fit(
      self,
      x: DataFrame,
      y: Series,
  ) -> 'XNB':
    """Calculate the best feature selection to be able to predict later.

    Args:
        x (DataFrame):  DataFrame containing the input features
        y (Series): Series containing the target variable

    Returns:
        XNB: Returns the instance itself.
    """
    class_values = set(y)
    bw_dict = self._calculate_bandwidth(
        x, y, self.bw_function, self.n_sample, class_values)
    kde_list = self._calculate_kdes(
        x, y, self.kernel, self.algorithm, bw_dict, self.n_sample, class_values)
    ranking = self._calculate_divergence(kde_list)
    self._calculate_feature_selection(ranking, class_values)
    self._calculate_necessary_kde(x, y, bw_dict, self.kernel, self.algorithm)
    self._calculate_target_representation(y, class_values)

    self.is_fitted_ = True
    return self

  def predict_proba(self, x: DataFrame) -> np.ndarray:
    """Return the probabilities of each class for all rows in the DataFrame.

    Args:
        x (DataFrame): DataFrame containing the input to predict probabilities.

    Returns:
        np.ndarray: Array where each row contains the probabilities for each
        class.
    """
    self._check_if_not_fitted_error()

    log_probs = self._calculate_class_log_probabilities(x)
    return self._normalize_probabilities(log_probs)

  def predict(self, x: DataFrame) -> np.ndarray:
    """Return the predicted class for each row in the DataFrame.

    Args:
        x (DataFrame): DataFrame containing the input to predict.

    Returns:
        np.ndarray: Numpy array containing the predicted class for each row.
    """
    self._check_if_not_fitted_error()

    log_probs = self._calculate_class_log_probabilities(x)
    probabilities = self.predict_proba(x)
    class_indices = np.argmax(probabilities, axis=1)
    return np.array(sorted(log_probs.keys()))[class_indices]

  def _check_if_not_fitted_error(self) -> None:
    cond1 = not hasattr(self, '_kernel_density_dict')
    cond2 = not hasattr(self, '_class_representation')
    cond3 = not hasattr(self, 'is_fitted_')

    if cond1 or cond2 or cond3:
      raise NotFittedError()

  def _calculate_bandwidth(
      self,
      x: DataFrame,
      y: Series,
      bw_function_name: str,
      n_sample: int,
      class_values: set
  ) -> Dict[Union[str, float], Dict[Union[str, float], float]]:
    bw_dict = {}
    bw_function = get_bandwidth_function(bw_function_name)

    p_len = len(class_values) * len(x.columns)
    p_title = 'Calculating bandwidths'
    with progress_bar(self.show_progress_bar, p_len, p_title) as next_bar:
      for class_value in class_values:
        bw_dict[class_value] = {}
        d = x[y == class_value]
        for feature in x.columns:
          bw = bw_function(d[feature], n_sample)
          bw_dict[class_value][feature] = bw
          next_bar()

    return bw_dict

  def _calculate_kdes(
      self,
      x: DataFrame,
      y: Series,
      kernel: str,
      algorithm: str,
      bw_dict: Dict[Union[str, float], Dict[Union[str, float], float]],
      n_sample: int,
      class_values: set
  ) -> List[KDE]:
    """Calculate the KDE for each class and feature."""
    kde_list = []

    p_len = len(class_values) * len(x.columns)
    p_title = 'Calculating all the KDEs'
    with progress_bar(self.show_progress_bar, p_len, p_title) as next_bar:
      for class_value in class_values:
        data_class = x[y == class_value]
        for feature in x.columns:
          # Calculate x_points
          data_var = x[feature]
          minimum, maximum = data_var.min(), data_var.max()
          x_points = np.linspace(minimum, maximum, n_sample)
          bw = bw_dict[class_value][feature]
          data = data_class[feature]

          # Fit Kernel Density
          kde = KernelDensity(
              kernel=kernel,
              bandwidth=bw,
              algorithm=algorithm
          ).fit(data.values[:, np.newaxis])

          # Calculate y_points
          y_points = np.exp(kde.score_samples(x_points[:, np.newaxis]))

          # Append the KDE object to the results list
          kde_list.append(KDE(
              feature,
              class_value,
              kde,
              x_points,
              y_points
          ))
          next_bar()

    return kde_list

  def _calculate_divergence(
      self,
      kde_list: List[KDE]
  ) -> DataFrame:
    """Calculate the divergence (distance) between each classes."""
    kde_dict = defaultdict(list[KDE])

    for kde in kde_list:
      kde_dict[kde.feature].append(kde)

    scores = []
    p_len = len(kde_dict)
    p_title = 'Calculating divergence'
    with progress_bar(self.show_progress_bar, p_len, p_title) as next_bar:
      for feature in kde_dict:
        for index_1 in range(len(kde_dict[feature]) - 1):
          t1 = kde_dict[feature][index_1].target_value
          f1 = kde_dict[feature][index_1].feature
          for index_2 in range(index_1 + 1, len(kde_dict[feature])):
            t2 = kde_dict[feature][index_2].target_value
            f2 = kde_dict[feature][index_2].feature
            if t1 != t2 and f1 == f2:
              p = self._normalize(kde_dict[feature][index_1].y_points)
              q = self._normalize(kde_dict[feature][index_2].y_points)
              hellinger = self._hellinger_distance(p, q)
              scores.append([f1, t1, t2, hellinger])
        next_bar()

    ranking = DataFrame(
        scores,
        columns=['feature', 'p0', 'p1', 'hellinger']
    ).drop_duplicates().sort_values(
        by=['hellinger', 'feature', 'p0', 'p1'],
        ascending=[False, True, True, True]
    )
    threshold = sum(ranking.hellinger) / len(ranking.hellinger)
    return ranking[ranking.hellinger > threshold].reset_index(drop=True)

  @staticmethod
  def _hellinger_distance(p: np.ndarray, q: np.ndarray) -> float:
    """Calculate the Hellinger distance between two distributions."""
    try:
      s = sum([sqrt(a * b) for a, b in zip(p, q)])
    except ValueError:
      s = 0.0
    s = max(0, min(1, s))
    return sqrt(1 - s)

  @staticmethod
  def _normalize(data: np.ndarray) -> np.ndarray:
    s = np.sum(data)
    if s > 0:
      data = data / s
    s = np.sum(data)
    while s > 1.0:
      data = data / s
      s = np.sum(data)
    return data

  def _calculate_feature_selection(
          self,
          ranking: DataFrame,
          class_values: set
  ) -> Dict[str, Set[Union[str, float]]]:
    """Calculate the feature selection for each class."""
    threshold = 0.999
    stop_dict = defaultdict(dict[str, bool])
    hellinger_dict = defaultdict(set)
    for cv_1, cv_2 in product(class_values, repeat=2):
      if cv_1 != cv_2:
        stop_dict[cv_1][cv_2] = False

    p_len = len(ranking)
    p_title = 'Calculating feature selection'
    with progress_bar(self.show_progress_bar, p_len, p_title) as next_bar:
      for _, row in ranking.iterrows():
        feature = row.feature
        class_1 = row.p0
        class_2 = row.p1
        hellinger = row.hellinger
        hellinger_dict, stop_dict = self._update_feature_selection_dict(
            hellinger_dict,
            stop_dict,
            feature,
            hellinger,
            threshold,
            class_a=class_1,
            class_b=class_2
        )
        hellinger_dict, stop_dict = self._update_feature_selection_dict(
            hellinger_dict,
            stop_dict,
            feature,
            hellinger,
            threshold,
            class_a=class_2,
            class_b=class_1
        )
        next_bar()

    self._feature_selection_dict = defaultdict(set[Union[str, float]])
    for class_value in hellinger_dict.keys():
      cv = (
          float(class_value)
          if isinstance(class_value, float)
          else class_value
      )
      self._feature_selection_dict[cv] = {
          float(cfd.feature)
          if isinstance(cfd.feature, float)
          else cfd.feature
          for cfd in hellinger_dict[cv]
      }
    self._feature_selection_dict = dict(self._feature_selection_dict)
    return self._feature_selection_dict

  @staticmethod
  def _update_feature_selection_dict(
          hellinger_dict: Dict[Union[str, float], Set[_ClassFeatureDistance]],
          stop_dict: Dict[str, Dict[str, bool]],
          feature: str,
          hellinger: float,
          threshold: float,
          class_a: str,
          class_b: str
  ) -> Tuple[
      Dict[Union[str, float], Set[_ClassFeatureDistance]],
      Dict[str, Dict[str, bool]]
  ]:
    """Auxiliary method to calculate the feature selection."""
    if not stop_dict[class_a][class_b]:
      not_in_dict = class_a not in {
          x.class_value for x in hellinger_dict[class_b]
      }
      if not_in_dict:
        stop_dict[class_a][class_b] = hellinger >= threshold
      else:
        p = 1
        for x in hellinger_dict[class_b]:
          if x.class_value == class_a:
            p *= (1 - x.distance)
        stop_dict[class_a][class_b] = (1 - (p * (1 - hellinger))) >= threshold

      hellinger_dict[class_b].add(
          _ClassFeatureDistance(
              class_value=class_a,
              feature=feature,
              distance=hellinger
          )
      )
    return hellinger_dict, stop_dict

  def _calculate_target_representation(
      self,
      target_col: Series,
      class_values: Set
  ) -> Dict:
    """Calculate the percentage representation for each class."""
    self._class_representation = {}
    for class_value in class_values:
      target_count = target_col.value_counts().get(class_value, 0)
      total_count = len(target_col)
      self._class_representation[class_value] = target_count / total_count
    return self._class_representation

  def _calculate_necessary_kde(
      self,
      x: DataFrame,
      y: Series,
      bw_dict: Dict[Union[str, float], Dict[Union[str, float], float]],
      kernel: str,
      algorithm: str
  ) -> Dict:
    """Calculate the KDE for each class."""
    self._kernel_density_dict = {}

    p_len = len(self._feature_selection_dict)
    p_title = 'Calculating only the necessary KDEs before the prediction'
    with progress_bar(self.show_progress_bar, p_len, p_title) as next_bar:
      for class_value, features in self._feature_selection_dict.items():
        data_class = x[y == class_value]
        self._kernel_density_dict[class_value] = {}
        for feature in features:
          data = data_class[feature]
          bw = bw_dict[class_value][feature]
          kde = KernelDensity(
              kernel=kernel,
              bandwidth=bw,
              algorithm=algorithm
          ).fit(data.values[:, np.newaxis])
          self._kernel_density_dict[class_value][feature] = kde
        next_bar()

    return self._kernel_density_dict

  def _calculate_class_log_probabilities(self, x: DataFrame) -> Dict:
    """Calculate the log probabilities for all classes for multiple rows.

    ## Args:
    :param x: DataFrame containing the input data.

    ## Returns:
    :return: Dictionary where each class is a key
    and the value is an array of log probabilities.
    """
    log_probs = {
        class_value: np.zeros(len(x))
        for class_value in self.feature_selection_dict
    }

    for class_value, features in self.feature_selection_dict.items():
      for feature in features:
          # Get KDE scores for the entire column
        log_probs[class_value] += self\
            ._kernel_density_dict[class_value][feature].score_samples(
            x[feature].values[:, np.newaxis]
        )
      # Add log prior probabilities
      log_probs[class_value] += np.log(self._class_representation[class_value])

    return log_probs

  def _normalize_probabilities(self, log_probs: Dict) -> np.ndarray:
    """Normalize log probabilities to return actual probabilities for all rows.

    ## Args:
    :param log_probs: Dictionary of log probabilities for each class.

    ## Returns:
    :return: Array of shape (n_samples, n_classes) with normalized probs.
    """
    log_probs_matrix = np.vstack(
        [log_probs[class_value] for class_value in sorted(log_probs.keys())]).T
    probs_matrix = np.exp(log_probs_matrix)
    probs_matrix /= probs_matrix.sum(axis=1, keepdims=True)
    return probs_matrix


class NotFittedError(ValueError, AttributeError):
  """Exception class to raise if estimator is used before fitting.

  This class inherits from both ValueError and AttributeError to help with
  exception handling and backward compatibility.
  """

  def __init__(
      self,
      message=(
          'This XNB instance is not fitted yet.',
          ' Call "fit" with appropriate arguments before using this estimator.'
      )
  ):
    """Initialize the error."""
    self.message = message

  def __str__(self):
    """Return the message."""
    return self.message



from sklearn.neighbors import KernelDensity
from numpy import ndarray


class KDE():
  """Class to represent a Kernel Density Estimation."""

  def __init__(
          self,
          feature: str,
          target_value: str,
          kernel_density: KernelDensity,
          x_points: ndarray,
          y_points: ndarray
  ) -> None:
    """Initialize the KDE object."""
    self.feature = feature
    self.target_value = target_value
    self.kernel_density = kernel_density
    self.x_points = x_points
    self.y_points = y_points



from alive_progress import alive_bar
from contextlib import contextmanager


class _DummyBar:
  def __call__(self):
    pass


@contextmanager
def progress_bar(enabled: bool, total: int, title: str = ''):
  """Context manager to show a progress bar."""
  if enabled:
    with alive_bar(total, title=title) as bar:
      yield bar
  else:
    yield _DummyBar()



"""Enum classes used in Explicable Naive Bayes."""
from enum import Enum


class BWFunctionName(str, Enum):
  """Enum class to select the bandwidth function."""

  HSILVERMAN = 'hsilverman'
  HSCOTT = 'hscott'
  HSJ = 'hsj'
  BEST_ESTIMATOR = 'best_estimator'

  def __str__(self) -> str:
    """Return the enum value."""
    return self.value


class Kernel(str, Enum):
  """Enum class to select the kernel function."""

  GAUSSIAN = 'gaussian'
  COSINE = 'cosine'
  EPANECHNIKOV = 'epanechnikov'
  TOPHAT = 'tophat'
  EXPONENTIAL = 'exponential'
  LINEAR = 'linear'

  def __str__(self) -> str:
    """Return the enum value."""
    return self.value


class Algorithm(str, Enum):
  """Enum class to select the algorithm."""

  KD_TREE = 'kd_tree'
  BALL_TREE = 'ball_tree'
  AUTO = 'auto'

  def __str__(self) -> str:
    """Return the enum value."""
    return self.value
